@@ -32,6 +32,8 @@ class TorchModel(ABC):
         self.module: torch.nn.Module = None
         self.outputScaler: Optional[TensorScaler] = None
         self.inputScaler: Optional[TensorScaler] = None
+        self.bestEpoch = None
+        self._gpu = None
 
     def setTorchModule(self, module: torch.nn.Module):
         self.module = module
@@ -57,14 +59,29 @@ class TorchModel(ABC):
     def _loadModel(self, modelFile):
         try:
             self.module = torch.load(modelFile)
+            self._gpu = self._getGPUFromModelParameterDevice()
         except:
             if self._isCudaEnabled():
-                self.log.warning("Loading of CUDA model failed, trying without CUDA...")
+                if torch.cuda.device_count() > 0:
+                    newDevice = "cuda:0"
+                else:
+                    newDevice = "cpu"
+                self.log.warning(f"Loading of CUDA model failed, trying to map model to device {newDevice}...")
                 if type(modelFile) != str:
                     modelFile.seek(0)
-                self.module = torch.load(modelFile, map_location='cpu')
-                self._setCudaEnabled(False)
-                self.log.info("Model successfully loaded to CPU")
+                try:
+                    self.module = torch.load(modelFile, map_location=newDevice)
+                except:
+                    self.log.warning(f"Failure to map model to device {newDevice}, trying CPU...")
+                    if newDevice != "cpu":
+                        newDevice = "cpu"
+                        self.module = torch.load(modelFile, map_location=newDevice)
+                if newDevice == "cpu":
+                    self._setCudaEnabled(False)
+                    self._gpu = None
+                else:
+                    self._gpu = 0
+                self.log.info(f"Model successfully loaded to {newDevice}")
             else:
                 raise
 
@@ -74,7 +91,7 @@ class TorchModel(ABC):
 
     def __getstate__(self):
         state = dict(self.__dict__)
-        del state["model"]
+        del state["module"]
         state["modelBytes"] = self.getModuleBytes()
         return state
 
@@ -124,6 +141,7 @@ class TorchModel(ABC):
             X = torch.from_numpy(X).float()
 
         if self._isCudaEnabled():
+            torch.cuda.set_device(self._gpu)
             X = X.cuda()
         if scaleInput:
             X = self.inputScaler.normalise(X)
@@ -166,6 +184,14 @@ class TorchModel(ABC):
             nnOptimiserParams["cuda"] = self.cuda
         optimiser = NNOptimiser(**nnOptimiserParams)
         optimiser.fit(self, data)
+        self.bestEpoch = optimiser.getBestEpoch()
+        self._gpu = self._getGPUFromModelParameterDevice()
+
+    def _getGPUFromModelParameterDevice(self) -> Optional[int]:
+        try:
+            return next(self.module.parameters()).get_device()
+        except:
+            return None
 
 
 class VectorTorchModel(TorchModel, ABC):
@@ -191,7 +217,7 @@ class VectorTorchModel(TorchModel, ABC):
 
 
 class TorchVectorRegressionModel(VectorRegressionModel):
-    def __init__(self, modelClass: Callable[..., VectorTorchModel], modelArgs=(), modelKwArgs=None,
+    def __init__(self, modelClass: Callable[..., TorchModel], modelArgs=(), modelKwArgs=None,
             normalisationMode=NormalisationMode.NONE, nnOptimiserParams=None):
         """
         :param modelClass: the constructor with which to create the wrapped torch vector model
@@ -212,19 +238,32 @@ class TorchVectorRegressionModel(VectorRegressionModel):
         self.modelClass = modelClass
         self.modelArgs = modelArgs
         self.modelKwArgs = modelKwArgs
-        self.model: Optional[VectorTorchModel] = None
+        self.model: Optional[TorchModel] = None
 
-    def _createWrappedTorchVectorModule(self) -> VectorTorchModel:
+    def _createWrappedTorchVectorModule(self) -> TorchModel:
         return self.modelClass(*self.modelArgs, **self.modelKwArgs)
 
+    def _createDataSetProvider(self, inputs: pd.DataFrame, outputs: pd.DataFrame) -> TorchDataSetProvider:
+        dataUtil = VectorDataUtil(inputs, outputs, self.model.cuda, normalisationMode=self.normalisationMode)
+        return TorchDataSetProviderFromDataUtil(dataUtil, self.model.cuda)
+
     def _fit(self, inputs: pd.DataFrame, outputs: pd.DataFrame):
         self.model = self._createWrappedTorchVectorModule()
-        dataUtil = VectorDataUtil(inputs, outputs, self.model.cuda, normalisationMode=self.normalisationMode)
-        dataSetProvider = TorchDataSetProviderFromDataUtil(dataUtil, self.model.cuda)
+        dataSetProvider = self._createDataSetProvider(inputs, outputs)
         self.model.fit(dataSetProvider, **self.nnOptimiserParams)
 
+    def _predictOutputsForInputDataFrame(self, inputs: pd.DataFrame) -> np.ndarray:
+        results = []
+        i = 0
+        batchSize = 2**13
+        while i < len(inputs):
+            inputSlice = inputs.iloc[i:i+batchSize]
+            results.append(self.model.applyScaled(inputSlice.values, asNumpy=True))
+            i += batchSize
+        return np.concatenate(results)
+
     def _predict(self, inputs: pd.DataFrame) -> pd.DataFrame:
-        yArray = self.model.applyScaled(inputs.values)
+        yArray = self._predictOutputsForInputDataFrame(inputs)
         return pd.DataFrame(yArray, columns=self.getModelOutputVariableNames())
 
     def __str__(self):
