@@ -4,6 +4,8 @@ import torch
 from torch import nn
 from torch.nn import functional as F
 
+from ..util.string import objectRepr
+
 
 class MCDropoutCapableNNModule(nn.Module, ABC):
     """
@@ -89,6 +91,11 @@ class MultiLayerPerceptron(MCDropoutCapableNNModule):
             self.layers.append(nn.Linear(prevDim, dim))
             prevDim = dim
 
+    def __str__(self):
+        return objectRepr(self, dict(inputDim=self.inputDim, outputDim=self.outputDim, hiddenDims=self.hiddenDims,
+            hidActivationFn=self.hidActivationFn.__name__ if self.hidActivationFn is not None else None, outputActivationFn=self.outputActivationFn.__name__ if self.outputActivationFn is not None else None,
+            pDropout=self.pDropout))
+
     def forward(self, x):
         for i, layer in enumerate(self.layers):
             isLast = i+1 == len(self.layers)
